import os
from os.path import join as opj
from datetime import datetime
import datetime
import urllib
from pathlib import Path
from typing import Union
import gc
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception, retry_if_exception_type
import requests
import random
import json
import pandas as pd
import time
import statistics
from progressbar import ProgressBar
import progressbar
import numpy as np
import string
import io
from scipy import interpolate
import copy
import requests
from datetime import datetime
import string
from io import StringIO
import io
import os
from scipy import interpolate
from scipy.interpolate import UnivariateSpline
import logging
from functools import lru_cache
from datetime import datetime
import pandas as pd
import numpy as np
from scipy import interpolate
import copy
import warnings

# we only need to retry if any of these error occured, the rest we dont need to know
NETWORK_EXCEPTIONS = (requests.ConnectionError, requests.Timeout, requests.RequestException)


def try_many_times(k_times):
    if k:
        return retry(stop=stop_after_attempt(k_times), wait=wait_fixed(0.5),
                     retry=retry_if_exception_type(NETWORK_EXCEPTIONS))


def generate_unique_name(base_name, length=6):
    # TODO this function has 4 duplicates
    random_suffix = ''.join(random.choices(string.ascii_lowercase, k=length))
    unique_name = base_name + '_' + random_suffix
    return unique_name


# from US_states_abbreviation import get_abbreviation

# let us manage network issues more gracefully
@retry(stop=stop_after_attempt(3), wait=wait_fixed(0.5), retry=retry_if_exception_type(NETWORK_EXCEPTIONS))
def get_iem_by_station(dates_tuple, station, path, met_tag):
    """
      ``dates_tuple``: (tuple, list) is a tupple/list of strings with date ranges
      
      - an example date string should look like this: ``dates`` = ["01-01-2012","12-31-2012"]
      ``station``: (str) is the station where toe xtract the data from
      -If ``station`` is given data will be downloaded directly from the station the default is false.
      
      :param met_tag: your preferred suffix to save on file

      """
    # access the elements in the metdate class above
    weather_dates = MetDate(dates_tuple)
    stationX = station[:2]
    state_clim = stationX + "CLIMATE"
    str0 = "http://mesonet.agron.iastate.edu/cgi-bin/request/coop.py?network="
    str1 = str0 + state_clim + "&stations=" + station
    str2 = str1 + "&year1=" + weather_dates.year_start + "&month1=" + weather_dates.start_month + "&day1=" + weather_dates.start_day + "&year2=" + weather_dates.year_end + "&month2=" + weather_dates.end_month + "&day2=" + weather_dates.end_day
    str3 = (str2 + "&vars%5B%5D=apsim&what=view&delim=comma&gis=no")
    url = str3
    rep = requests.get(url)
    if rep.ok:
        met_name_of_file = station + met_tag + ".met"
        os.chdir(path)
        if not os.path.exists('weatherdata'):
            os.mkdir('weatherdata')
        pt = os.path.join('weatherdata', met_name_of_file)

        with open(pt, 'wb') as met_fileX:
            # remember to include the file extension name
            met_fileX.write(rep.content)
            rep.close()
            met_fileX.close()
            print(rep.content)
    else:
        print("Failed to download the data web request returned code: ", rep)


class MetDate:
    def __init__(self, dates):
        self.start_date = dates[0]
        self.end_date = dates[1]
        self.start_month = dates[0][:2]
        self.end_month = dates[1][:2]
        self.year_start = dates[0].split("-")[2]
        self.year_end = dates[1].split("-")[2]
        import datetime
        self.start_day = datetime.datetime.strptime(dates[0], '%m-%d-%Y').strftime('%j')
        self.end_day = datetime.datetime.strptime(dates[1], '%m-%d-%Y').strftime('%j')


dates = ['01-01-2000', '12-31-2020']

states = {
    'AK': 'Alaska',
    'AL': 'Alabama',
    'AR': 'Arkansas',
    'AZ': 'Arizona',
    'CA': 'California',
    'CO': 'Colorado',
    'CT': 'Connecticut',
    'DC': 'District of Columbia',
    'DE': 'Delaware',
    'FL': 'Florida',
    'GA': 'Georgia',
    'HI': 'Hawaii',
    'IA': 'Iowa',
    'ID': 'Idaho',
    'IL': 'Illinois',
    'IN': 'Indiana',
    'KS': 'Kansas',
    'KY': 'Kentucky',
    'LA': 'Louisiana',
    'MA': 'Massachusetts',
    'MD': 'Maryland',
    'ME': 'Maine',
    'MI': 'Michigan',
    'MN': 'Minnesota',
    'MO': 'Missouri',
    'MS': 'Mississippi',
    'MT': 'Montana',
    'NC': 'North Carolina',
    'ND': 'North Dakota',
    'NE': 'Nebraska',
    'NH': 'New Hampshire',
    'NJ': 'New Jersey',
    'NM': 'New Mexico',
    'NV': 'Nevada',
    'NY': 'New York',
    'OH': 'Ohio',
    'OK': 'Oklahoma',
    'OR': 'Oregon',
    'PA': 'Pennsylvania',
    'RI': 'Rhode Island',
    'SC': 'South Carolina',
    'SD': 'South Dakota',
    'TN': 'Tennessee',
    'TX': 'Texas',
    'UT': 'Utah',
    'VA': 'Virginia',
    'VT': 'Vermont',
    'WA': 'Washington',
    'WI': 'Wisconsin',
    'WV': 'West Virginia',
    'WY': 'Wyoming'
}

# flip the keys
new_dict = {}
for k, v in states.items():
    new_dict[v] = k


def get_abbreviation(x):
    ab = new_dict[x]
    return (ab)


# function to define the date ranges
def daterange(start, end):
    """
  :param start: (int) the starting year to download the weather data
  -----------------
  :param end: (int) the year under which download should stop
  """
    startdates = '01-01'
    enddates = '12-31'
    end = str(end) + "-" + enddates
    start = str(start) + "-" + startdates
    drange = pd.date_range(start, end)
    return (drange)


# check if a year is aleap year
def isleapyear(year):
    if (year % 400 == 0) and (year % 100 == 0) or (year % 4 == 0) and (year % 100 != 0):
        return (True)
    else:
        return (False)


# download radiation data for replacement
@retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(NETWORK_EXCEPTIONS))
def get_nasarad(lonlat, start, end):
    lon = lonlat[0]
    lat = lonlat[1]
    pars = "ALLSKY_SFC_SW_DWN"
    rm = f'https://power.larc.nasa.gov/api/temporal/daily/point?start={start}0101&end={end}1231&latitude={lat}&longitude={lon}&community=ag&parameters={pars}&format=json&user=richard&header=true&time-standard=lst'
    data = requests.get(rm)
    dt = json.loads(data.content)
    df = pd.DataFrame(dt["properties"]['parameter'])
    if len(df) == len(daterange(start, end)):
        return df

    # fucntion to download data from daymet

@lru_cache(maxsize=500)
def get_met_from_day_met(lonlat: Union[tuple, list, np.ndarray], start: int,
                         end: int, filename: str,
                         fill_method: str = 'ffill',
                         retry_number: Union[int, None] = 1, **kwa: None) -> str:
    """
    Collect weather from daymet solar radiation is replaced with that of nasapower API

    Parameters
    ---------------

    ``lonlat``:
         tuple, list, np.ndarray.

    ``retry_number``:
        (int): retry number of times in case of network errors.

    ``filename``.
         met file name to save on disk.

    ``start``.
         Starting year of the met data.

    ``end``.
         Ending year of the met data.

    ``lonlat``.
         (tuple, list, array): A tuple of XY cordnates, longitude first, then latitude second.

    ``fill_method``.
         (str, optional): fills the missing data based pandas fillna method arguments may be bfill, ffill defaults to ffill.

    ``keyword``.
         ``timeout`` specifies the waiting time.

        ``wait``: the time in secods to try for every retry in case of network errors.

    ``returns``
       A complete path to the new met file but also write the met file to the disk in the working dir_path.

    Example:
    --------------

          >>> from apsimNGpy.manager.weathermanager import get_met_from_day_met
          >>> wf = get_met_from_day_met(lonlat=(-93.04, 42.01247),
          >>> start=2000, end=2020,timeout = 30, wait =2, retry_number=3, filename='daymet.met')

    """
    _start = f"{start}-01-01"
    _end = f'{end}-12-31'
    full_date_range = pd.date_range(start=_start, end=_end, freq='D')
    if start < 1980 or end > 2023:
        print("requested year precedes valid data range! \n"
              " end years should not exceed 2021 and start year should not be less than 1980")
    else:
        base_url = 'https://daymet.ornl.gov/single-pixel/api/data?'
        lat_str, lon_str = 'lat=' + str(lonlat[1]), '&lon=' + str(lonlat[0])

        var_headers = ['dayl', 'prcp', 'srad', 'tmax', 'tmin', 'vp', 'swe']
        years_in_range = ",".join([str(year) for year in full_date_range.year.unique()])
        # join the years as a string
        years_str, var_field = "&years=" + years_in_range,  ",".join(var_headers)

        var_str = "&measuredParams=" + var_field
        # join the string url together
        url = base_url + lat_str + lon_str + var_str + years_str

        def connect():
            """
            just to allow users enter number retries
            Returns
            -------

            """
            try:
                _conn = requests.get(url, timeout=kwa.get('timeout', 50))
                return _conn
            # We want to retry only if the network exceptions defined above occur not value errors or child errors and
            # so forth
            except NETWORK_EXCEPTIONS:
                raise

        if retry_number:
            # Apply the retry decorator to the connect function
            connect_with_retry = retry(wait=wait_fixed(kwa.get('wait', 0.5)),
                                       stop=stop_after_attempt(retry_number),
                                       )(connect)

            connector = connect_with_retry()
        else:
            connector = connect()

        text_str = connector.content
        connector.close()
        # Create an in-memory binary stream
        text_stream = io.BytesIO(text_str)
        # Read the data into a DataFrame
        day_met_read = pd.read_csv(text_stream, delimiter=',', skiprows=6)
        vp = day_met_read['vp (Pa)'] * 0.01
        # calculate radiation
        radiation = day_met_read['dayl (s)'] * day_met_read['srad (W/m^2)'] * 1e-06
        # re-arrange data frame
        flip_cat = {
            'year': 'year',
            'yday': 'day',
            'dayl (s)': 'radn',
            'tmax (deg c)': 'maxt',
            'tmin (deg c)': 'mint',
            'prcp (mm/day)': 'rain',
            'vp (Pa)': 'vp',
            'swe (kg/m^2)': 'swe'
        }
        df = day_met_read.rename(columns=flip_cat)
        df['radn'] = radiation
        df['vp'] = vp
        df.drop('srad (W/m^2)', axis=1, inplace=True)
        COLUMNs = ['year', 'day', 'radn', 'maxt', 'mint', 'rain', 'swe', 'vp']
        _data_frame = df[COLUMNs].copy()

        _data_frame['Date'] = pd.to_datetime(_data_frame.year.astype(str) + _data_frame.day.astype(str), format='%Y%j')
        data_to = _data_frame.set_index('Date')
        # now reindex according to the date range
        new_met = data_to.reindex(full_date_range, columns=COLUMNs)
        # replacements for year and data
        new_met['year'] = full_date_range.year.astype(int)
        new_met['day'] = full_date_range.dayofyear.astype('int')
        # Now reindexing does not change NAs that were already there, even if we pass a fill method to it,
        # so we need to change them here at once
        new_met = getattr(new_met, fill_method)()
        # replace radiation data
        rad = get_nasarad(lonlat, start, end)
        new_met["radn"] = rad.ALLSKY_SFC_SW_DWN.values
        mean_max_temp = new_met['maxt'].mean(skipna=True, numeric_only=None)
        mean_mint = new_met['mint'].mean(skipna=True, numeric_only=None)
        AMP = round(mean_max_temp - mean_mint, 2)
        tav = round(statistics.mean((mean_max_temp, mean_mint)), 2)
        tile = connector.headers["Content-Disposition"].split("=")[1].split("_")[0]
        short_file_name = filename or "Daymet" + '.met'
        if not os.path.exists('weatherdata'):
            os.makedirs('weatherdata')
        file_name_path = os.path.join('weatherdata', short_file_name)
        headers = ['year', 'day', 'radn', 'maxt', 'mint', 'rain', 'vp', 'swe']
        header_string = " ".join(headers) + "\n"
        # close and append new lines
        with open(file_name_path, "a") as f2app:
            f2app.writelines([f'!site: {tile}\n', f'latitude = {lonlat[1]} \n', f'longitude = {lonlat[0]}\n',
                              f'tav ={tav}\n', f'amp ={AMP}\n'])
            f2app.writelines([header_string])
            f2app.writelines(['() () (MJ/m2/day) (oC) (oC) (mm) (hPa) (kg/m2)\n'])
            # append the weather data
            data_rows = new_met.apply(lambda row: " ".join([str(row[header]) for header in headers]) + '\n', axis=1)
            # Write all rows to the file at once
            f2app.writelines(data_rows)
        # we need a clean up
        del new_met, _data_frame, headers, mean_max_temp, data_to, text_str, radiation, years_in_range, text_stream
        return file_name_path


# download weather data from nasa power
def get_nasa(lonlat, start, end):
    lon = lonlat[0]
    lat = lonlat[1]
    param = ["T2M_MAX", "T2M_MIN", "ALLSKY_SFC_SW_DWN", "PRECTOTCORR", "RH2M", "WS2M"]
    pars = ",".join(param)
    rm = f'https://power.larc.nasa.gov/api/temporal/daily/point?start={start}0101&end={end}1231&latitude={lat}&longitude={lon}&community=ag&parameters={pars}&format=json&user=richard&header=true&time-standard=lst'
    data = requests.get(rm)
    dt = json.loads(data.content)
    df = pd.DataFrame(dt["properties"]['parameter'])
    return df


def calculate_tav_amp(df):
    mean_maxt = df['maxt'].mean(skipna=True, numeric_only=None)
    mean_mint = df['mint'].mean(skipna=True, numeric_only=None)
    AMP = round(mean_maxt - mean_mint, 2)
    tav = round(np.mean([mean_maxt, mean_mint]), 3)
    return tav, AMP


def _create_met_header(met_file_name, lonlat, tav, AMP, site=None):
    if not site:
        site = 'Not stated'
    if os.path.isfile(met_file_name):
        os.remove(met_file_name)
    headers = ['year', 'day', 'radn', 'maxt', 'mint', 'rain']
    header_string = " ".join(headers) + "\n"
    # close and append new lines
    with open(met_file_name, "a") as f2app:
        f2app.writelines(
            [f'!site: {site}\n', f'latitude = {lonlat[1]} \n', f'longitude = {lonlat[0]}\n', f'tav ={tav}\n',
             f'amp ={AMP}\n'])
        f2app.writelines([header_string])
        f2app.writelines(['() () (MJ/m2/day) (oC) (oC) (mm)\n'])

@lru_cache(maxsize=50)
def impute_data(met, method="mean", verbose=False, **kwargs):
    """
    Imputes missing data in a pandas DataFrame using specified interpolation or mean value.

    Parameters:
    _______________________

    ``met``: (pd.DataFrame): DataFrame with missing values.

    ``method``: (str, optional): Method for imputing missing values ("approx", "spline", "mean"). Default is "mean".

    ``verbose``: (bool, optional): If True, prints detailed information about the imputation. Default is False.

    - **kwargs (dict, optional): Additional keyword arguments including 'copy' (bool) to deep copy the DataFrame.

    ``Returns:``
       - ``pd.DataFrame``: DataFrame with imputed missing values.
    """

    if kwargs.get('copy', False):
        met = copy.deepcopy(met)

    if not isinstance(met, pd.DataFrame):
        raise ValueError("The 'met' parameter must be a pandas DataFrame.")

    valid_methods = ["approx", "spline", "mean"]
    if method not in valid_methods:
        raise ValueError(f"Invalid method '{method}'. Valid options are {valid_methods}.")

    def vprint(*args):
        """Conditional print based on verbose flag."""
        if verbose:
            print(*args)

    # Impute values based on the specified method
    for col_name in met.columns:
        if met[col_name].isna().any():
            vprint(f"Imputing missing values in column '{col_name}' using '{method}' method.")
            if method in ["approx", "spline"]:
                # Prepare for interpolation
                valid = met[~met[col_name].isna()]
                indices = np.arange(len(met))
                if method == "approx":
                    interp_func = interpolate.interp1d(valid.index, valid[col_name], bounds_error=False,
                                                       fill_value="extrapolate")
                else:  # spline
                    interp_func = interpolate.UnivariateSpline(valid.index, valid[col_name], s=0, ext=3)
                imputed_values = interp_func(indices)
                met[col_name] = imputed_values
            else:  # mean
                mean_value = met[col_name].mean(skipna=True)
                met[col_name].fillna(mean_value, inplace=True)

    return met


def separate_date(date_str):
    # Ensure the date string is of the correct format

    # Extracting year, month, and day
    year = date_str[:4]
    month = date_str[4:6]
    day = date_str[6:]

    return year, month, day


@retry(stop=stop_after_attempt(2), wait=wait_fixed(0.5), retry=retry_if_exception_type(NETWORK_EXCEPTIONS))
def get_nasa_data(lonlat, start, end):
    lon = lonlat[0]
    lat = lonlat[1]
    param = ["T2M_MAX", "T2M_MIN", "ALLSKY_SFC_SW_DWN", "PRECTOTCORR", "RH2M", "WS2M"]
    # dictionary
    """
    "PRECTOTCORR": Corrected Total Precipitation
    "WS2M" : wind speed at 2 meters
    RH2M: Relative Humidity at 2 Meters
    ALLSKY_SFC_SW_DWN: All Sky Surface Shortwave Downward Irradiance
    """
    pars = ",".join(param)
    rm = f'https://power.larc.nasa.gov/api/temporal/daily/point?start={start}0101&end={end}1231&latitude={lat}&longitude={lon}&community=ag&parameters={pars}&format=json&user=richard&header=true&time-standard=lst'
    data = requests.get(rm)
    dt = json.loads(data.content)
    fd = {}
    headers = ['year', 'day', 'radn', 'maxt', 'mint', 'rain', 'vp', 'swe']
    df = pd.DataFrame(dt["properties"]['parameter'])
    fd['year'] = [date_str[:4] for date_str in df.index]
    date_object = [datetime.strptime(date_str, '%Y%m%d') for date_str in df.index]
    fd['day'] = [date_obj.timetuple().tm_yday for date_obj in date_object]
    fd['radn'] = df.ALLSKY_SFC_SW_DWN
    fd['maxt'] = df.T2M_MAX
    fd['mint'] = df.T2M_MIN
    fd['rain'] = df.PRECTOTCORR
    df = pd.DataFrame(fd)
    return df


@lru_cache(maxsize=700) # since it is returning just strings no problem caching it
def get_met_nasa_power(lonlat, start=1990, end=2000, fname='get_met_nasa_power.met', site=None):
    df = get_nasa_data(lonlat, start, end)
    tav, AMP = calculate_tav_amp(df)
    _create_met_header(fname, lonlat, tav, AMP, site=site)
    data_rows = []
    headers = ['year', 'day', 'radn', 'maxt', 'mint', 'rain']
    for index, row in df.iterrows():
        current_row = []
        for header in headers:
            current_row.append(str(row[header]))
        current_str = " ".join(current_row) + '\n'
        data_rows.append(current_str)
    with open(fname, "a") as f2app:
        f2app.writelines(data_rows)
    return fname

@lru_cache(maxsize=500)
def _is_within_USA_mainland(lonlat):
    lon_min, lon_max = -125.0, -66.9  # Approximate longitudes for the west and east coasts
    lat_min, lat_max = 24.4, 49.4  # Approximate latitudes for the southern and northern borders

    # Check if given coordinates are within the bounding box
    if lon_min <= lonlat[0] <= lon_max and lat_min <= lonlat[1] <= lat_max:
        return True
    else:
        return False

@lru_cache(maxsize=600, typed=True)
def get_weather(lonlat:Union[tuple, list], start:int=1990, end:int=2020, source:str='daymet', filename:str='__met_.met'):
    """
        Collects data from various sources. This function is cached for speed, so when it is called multiple times with the same arguments during optimization, it will be very fast fast

        Only ``nasapower`` and ``dayment`` are currently supported sources, so it will raise an error if mesonnet is suggested.

        -Note if you are not in mainland USA, please don't pass source = ``'dayment'`` as it will raise an error due to geographical
             scope

         Parameters
         -----------------------

         ``lonlat``: (``tuple``) lonlat values

         ``start``: (``int``) start year

         ``end``: (``int``) end year

         ``source``: (``str``) source API for weather data

         ``filename``: (``str``) filename for saving on disk

        Example.

            >>> from apsimNGpy.manager.weathermanager import get_weather
            >>> from apsimNGpy.core.base_data import load_default_simulations

            # We are going to collect data from my hometown Kampala

            >>> kampala_loc = 35.582520, 0.347596

            # Notice it return a path to the downloaded weather file

            >>> met_file = get_weather(kampala_loc, start=1990, end=2020, source='nasa', filename='kampala_new.met')
            >>> print(met_file)

            # next we can pass this weather file to apsim model

            >>> maize_model = load_default_simulations(crop = 'maize')
            >>> maize_model.replace_met_file(weather_file = met_file)

    """
    if source not in ['nasa', 'daymet']:
        raise NotImplementedError('source must be either "nasa" or "daymet"')

    if source == 'daymet' and _is_within_USA_mainland(lonlat):
        file_name =  filename
        return get_met_from_day_met(lonlat, start=start, end=end, filename=file_name)
    elif source == 'nasa':
        file_name = filename
        return get_met_nasa_power(lonlat, start, end, fname=file_name)
    else:
        raise ValueError(
            f"Invalid source: {source} according to supplied {lonlat} lon_lat values try 'nasa' instead")


def read_apsim_met(met_path, skip=5, index_drop=0, separator=' '):
    try:
        data = pd.read_csv(met_path, skiprows=skip, delim_whitespace=True)
        return data.drop(0)

    except Exception as e:
        print(repr(e))


def write_edited_met(old: Union[str, Path], daf: pd.DataFrame, filename: str = "edited_met.met") -> str:

    existing_lines = []
    with open(old, 'r+') as file:
        for i, line in enumerate(file):
            existing_lines.append(line)
            if 'MJ' in line and 'mm' in line and 'day in line':
                print(line)
                break
    # iterate through the edited data frame met
    headers = daf.columns.to_list()
    for index, row in daf.iterrows():
        current_row = []
        for header in headers:
            current_row.append(str(row[header]))
        current_str = " ".join(current_row) + '\n'
        existing_lines.append(current_str)
    fname = os.path.join(os.path.dirname(old), filename)
    if os.path.exists(fname):
        os.remove(fname)
    with open(fname, "a") as df_2met:
        df_2met.writelines(existing_lines)
    return fname


def merge_columns(df1_main, common_column, df2, fill_column, df2_colummn):
    """
    Parameters:
    ``df_main`` (pd.DataFrame): The first DataFrame to be merged and updated.

    ``common_column`` (str): The name of the common column used for merging.

    ``df2`` (pd.DataFrame): The second DataFrame to be merged with 'df_main'.

    ``fill_column`` (str): The column in 'edit' to be updated with values from 'df2_column'.

    ``df2_column`` (str): The column in 'df2' that provides replacement values for 'fill_column'.

    ``Returns``:
      ``pd.DataFrame``: A new DataFrame resulting from the merge and update operations.
    """
    try:
        df = df1_main.merge(df2, on=common_column, how='left')
        dm_copy = copy.deepcopy(df)
        dm_copy[df2_colummn].fillna(dm_copy[fill_column], inplace=True)
        dm_copy[fill_column] = dm_copy[df2_colummn]
        # dm_copy.drop(columns=[df2_colummn], inplace=True)
        dm_copy.reset_index()
        return dm_copy
    except Exception as e:
        print(repr(e))


def validate_met(met):
    import datetime
    if not isinstance(met, pd.DataFrame):
        raise ValueError("object should be of class 'DataFrame'")

    if met.empty:
        raise ValueError("No rows of data present in this object.")

    expected_colnames = ['year', 'day', 'radn', 'maxt', 'mint', 'rain']  # Add more as per your actual data structure
    if len(met.columns) != len(expected_colnames) or not all(met.columns == expected_colnames):
        print("Names in DataFrame:", met.columns)
        logging.warning("Expected column names", expected_colnames)
        warnings.warn("Number of columns in DataFrame does not match expected column names")
    for cols in met.columns:
        if cols == 'year':
            try:
                met = met[met['year'] != '!site:']
                met[cols] = met[cols].astype(int)
            except:
                met[cols] = met[cols].astype('float')
        else:
            try:
                met[cols] = met[cols].astype('float')
            except:
                pass
    if met['year'].isna().any():
        logging.warning(met[met['year'].isna()])
        warnings.warn("Missing values found for year")

    if met['year'].min() < 1900:
        logging.warning(met[met['year'] < 1900])
        warnings.warn(" minimum year is less than 1500")

    if met['year'].max() > 2024:
        logging.warning(met[met['year'] > 2024])
        warnings.warn("year is greater than 2024")

    # Add similar checks for other columns like 'day', 'mint', 'maxt', 'radn', 'rain'
    # Implement any specific checks as in your R function
    # Example for 'day':
    if met['day'].isna().any():
        print(met[met['day'].isna()])
        warnings.warn("Missing values found for day")

    if met['day'].min() < 1:
        print(met[met['day'] < 1])
        warnings.warn(" met starting day is less than 1")

    if met['day'].max() > 366:
        print(met[met['day'] > 366])
        warnings.warn("day is greater than 366")

    # Example for date checks
    if len(met) > 0:
        first_day = datetime.datetime(int(met.iloc[0]['year']), 1, 1) + pd.to_timedelta(met.iloc[0]['day'] - 1,
                                                                                        unit='d')
        last_day = datetime.datetime(int(met.iloc[-1]['year']), 1, 1) + pd.to_timedelta(met.iloc[-1]['day'] - 1,
                                                                                        unit='d')
        expected_dates = pd.date_range(start=first_day, end=last_day, freq='D')
        if len(met) < len(expected_dates):
            warnings.warn(f"{len(expected_dates) - len(met)} date discontinuities found. Consider checking data.")
    check_rain = met[met['rain'] > 95]
    rad_high = met[met['radn'] > 41]
    if not rad_high.empty:
        print(rad_high)
        warnings.warn("radiation is too high")
    if not check_rain.empty:
        warnings.warn('probably rain is too high')
        print(met[met['rain'] > 95])

    def check_missing(col):
        if met[col].isna().any():
            warnings.warn(f"{col}: has missing values_****___")
            print(met[col].isna())

    check_radn = met[met['radn'] < 0]
    if not check_radn.empty:
        logging.warning('probably radiation is too low or missing')
        print(met[met['radn'] < 0])

    # Temperature checks could be similar, comparing 'maxt' and 'mint'
    # Radiation, rain, and other checks as per the original function's logic
    for cols in met.columns:
        check_missing(cols)
    min_t = met[met['mint'] < -60]
    if not min_t.empty:
        warnings.warn("min temp is too low")
        print(min_t)

    min_t_high = met[met['mint'] > 40]
    if not min_t_high.empty:
        warnings.warn('Minimum temp is too high')
        print(min_t_high)
    maxt = met[met['maxt'] < -60]
    if not maxt.empty:
        warnings.warn("max temp is too low")
        print(maxt)
    maxt_high = met[met['maxt'] > 60]
    if not maxt_high.empty:
        print("max temp is too high")
        print(maxt_high)
    if len(daterange(met.year.min(), met.year.max())) != len(met):
        warnings.warn('daterange is not standard')
    else:
        print('daterange matched expected')


from datetime import datetime, timedelta


def day_of_year_to_date(year, day_of_year):
    """
    Convert day of the year to a date.

    Parameters:
    -----------
    ``year`` : int
        The year to which the day of the year belongs.

    ``day_of_year`` : int
        The day of the year (1 to 365 or 366).

    ``Returns:``
    --------
    ``datetime.date`` : he corresponding date. ``datetime.date``
        T
    """
    return datetime(year, 1, 1) + timedelta(days=day_of_year - 1)

def impute_missing_leaps(dmet, fill=0):
    dmet['year'] = dmet['year'].astype(int)
    dmet['day'] = dmet['day'].astype(int)
    datee = [day_of_year_to_date(_year, day) for _year, day in zip(dmet.year, dmet.day)]

    # Create a copy and set the date as the index
    dmett = dmet.copy()
    dmett['datee'] = pd.to_datetime(datee)
    dmett.set_index('datee', inplace=True)
    indexDate = pd.date_range(start=dmett.index.min(), end=dmett.index.max(), freq='D')
    check = daterange(dmett.year.min(), dmett.year.max())
    # Reindex the DataFrame to include all dates in the range
    df_daily = dmett.reindex(check)
    df_fill = df_daily.fillna(fill).copy(deep=True)
    df_fill['day'] = df_fill.index.day
    df_fill['year'] = df_fill.index.year
    # Optionally, fill missing values (for example, with the mean of the column)

    return df_fill


gc.collect()
if __name__ == '__main__':
    from pathlib import Path

    p = Path.cwd().glob("*.met")
    [os.remove(i) for i in p]
    os.chdir(Path.home())
    import cProfile

    profiler = cProfile.Profile()
    profiler.enable()
    Name = "dumu.met"
    if os.path.exists(Name):
        os.remove(Name)
    profiler.disable()
    # profiler.print_stats(sort='time')
    a = time.perf_counter()
    xp = get_met_from_day_met(lonlat=(-93.50456, 42.601247), start=1989, end=2001, filename=Name, retry_number=3)

    b = time.perf_counter()
    print(b - a)
    from apsimNGpy.core.base_data import load_default_simulations

    model = load_default_simulations(crop='maize')
    # model.path = 'clone.apsimx'

    model.replace_met_file(weather_file=xp)
    model.run()



import os

from concurrent.futures import ProcessPoolExecutor, as_completed, ThreadPoolExecutor
from multiprocessing import cpu_count
from time import perf_counter
from apsimNGpy.settings import NUM_CORES
from apsimNGpy.core_utils.run_utils import run_model
from tqdm import tqdm
from apsimNGpy import settings
from apsimNGpy.core_utils.utils import select_process
from apsimNGpy.core_utils.database_utils import read_db_table
from apsimNGpy.parallel.safe import download_soil_table
from itertools import islice
import pandas as pd
import multiprocessing as mp
import types
from typing import Iterable

CPU = int(int(cpu_count()) * 0.5)
CORES = NUM_CORES


def select_type(use_thread: bool, n_cores: int):
    return ThreadPoolExecutor(n_cores) if use_thread else ProcessPoolExecutor(n_cores)


def custom_parallel(func, iterable: Iterable,  *args, **kwargs):
    """
    Run a function in parallel using threads or processes.

    *Args:
        ``func`` (callable): The function to ``run`` in parallel.

        ``iterable`` (iterable): An iterable of items that will be ran_ok by the function.

        ``*args``: Additional arguments to pass to the ``func`` function.

    Yields:
        Any: The results of the ``func`` function for each item in the iterable.

   **kwargs
     ``use_thread`` (bool, optional): If True, use threads for parallel execution; if False, use processes. Default is False.

     ``ncores`` (int, optional): The number of threads or processes to use for parallel execution. Default is 50% of cpu
         cores on the machine.

     ``verbose`` (bool): if progress should be printed on the screen, default is True.
         progress_message (str) sentence to display progress such processing weather please wait. defaults to f"Processing multiple jobs via 'func.__name__' please wait!".

     ``void`` (bool, optional): if True, it implies that the we start consuming data internally right away, recomended for methods that operates on objects without returning data,
         such that you dont need to unzip or iterate on such returned data objects.

    """

    use_thread, cpu_cores = kwargs.get('use_thread', False), kwargs.get('ncores', CORES)
    progress_message = kwargs.get('progress_message', f"Processing multiple jobs via '{func.__name__}' please wait!")
    progress_message += ": "
    void = kwargs.get('void', False)
    selection = select_type(use_thread=use_thread,
                            n_cores=cpu_cores)
    bar_format = f"{progress_message}{{l_bar}}{{bar}}| jobs completed: {{n_fmt}}/{{total_fmt}}| Elapsed time: {{elapsed}}"
    with selection as pool:
        futures = [pool.submit(func, i, *args) for i in iterable]

        # progress = tqdm(total=len(futures), position=0, leave=True,
        #                 bar_format=f'{progress_message} {|{bar}|}:' '{percentage:3.0f}% completed')
        with selection as pool:
            futures = [pool.submit(func, i, *args) for i in iterable]

            with tqdm(
                    total=len(futures),
                    position=0,
                    leave=True,
                    bar_format=bar_format
            ) as progress:

                if not void:
                    for future in as_completed(futures):
                        yield future.result()
                        progress.update(1)
                else:
                    for future in as_completed(futures):
                        future.result()  # discard result, just execute
                        progress.update(1)
                    return None




# _______________________________________________________________
def run_apsimx_files_in_parallel(iterable_files: Iterable, **kwargs):
    """
    Run APSIMX simulation from multiple files in parallel.

    Args:
    ``iterable_files`` (list): A list of APSIMX  files to be run in parallel.

    ``ncores`` (int, optional): The number of CPU cores or threads to use for parallel processing. If not provided, it defaults to 50% of available CPU cores.

    ``use_threads`` (bool, optional): If set to True, the function uses thread pool execution; otherwise, it uses process pool execution. Default is False.

    Returns:
    ``returns`` a generator object containing the path to the datastore or sql databases

    Example:

    # Example usage of read_result_in_parallel function

    >>> from apsimNgpy.parallel.process import run_apsimxfiles_in_parallel
    >>> simulation_files = ["file1.apsimx", "file2.apsimx", ...]  # Replace with actual database file names

    # Using processes for parallel execution

    >>> result_generator = run_apsimxfiles_in_parallel(simulation_files, ncores=4, use_threads=False)
    ```

    Notes:
    - This function efficiently reads db file results in parallel.
    - The choice of thread or process execution can be specified with the `use_threads` parameter.
    - By default, the function uses 50% of available CPU cores or threads if `ncores` is not provided.
    - Progress information is displayed during execution.
    - Handle any exceptions that may occur during execution for robust processing.
    """
    # remove duplicates. because duplicates will be susceptible to race conditioning in paralell computing
    Ncores = kwargs.get('ncores')
    if Ncores:
        ncores_2use = Ncores
    else:
        ncores_2use = int(cpu_count() * 0.50)

    return custom_parallel(run_model, iterable_files, ncores=ncores_2use, use_threads=kwargs.get('use_threads'))


def _read_result_in_parallel(iterable_files: Iterable, ncores: int = None, use_threads: bool = False,
                            report_name: str = "Report", **kwargs):
    """

    Read APSIMX simulation databases results from multiple files in parallel.

    Args:
    ``iterable_files`` (list): A list of APSIMX db files to be read in parallel.

    ``ncores`` (int, optional): The number of CPU cores or threads to use for parallel processing. If not provided, it defaults to 50% of available CPU cores.

    ``use_threads`` (bool, optional): If set to True, the function uses thread pool execution; otherwise, it uses process pool execution. Default is False.

    ``report_name`` the name of the report table defaults to "Report" you can use None to return all

    Returns:
    - ``generator``: A generator yielding the simulation data read from each file.

    Example:

    # Example usage of read_result_in_parallel function

    >>> from  apsimNgpy.parallel.process import read_result_in_parallel

    >>> simulation_files = ["file1.db", "file2.db", ...]  # Replace with actual database file names

    # Using processes for parallel execution
    result_generator = read_result_in_parallel(simulation_files, ncores=4, use_threads=False)

    # Iterate through the generator to process results
    for data in result_generator:
        print(data)
    it depends on the child of data but pd.concat could be a good option on the returned generator
    Kwargs
        func custom method for reading data
    ```

    Notes:
    - This function efficiently reads db file results in parallel.
    - The choice of thread or process execution can be specified with the `use_threads` parameter.
    - By default, the function uses 50% of available CPU cores or threads if `ncores` is not provided.
    - Progress information is displayed during execution.
    - Handle any exceptions that may occur during execution for robust processing.
    """

    func = kwargs.get('func', None)
    progress_msg = 'Reading data from path: '
    Ncores = ncores
    if Ncores:
        ncores_2use = Ncores
    else:
        ncores_2use = int(cpu_count() * 0.50)
    worker = func or read_db_table
    return custom_parallel(worker, iterable_files, report_name, ncores=ncores_2use, progress_message=progress_msg,
                           use_threads=use_threads)


def download_soil_tables(iterable: Iterable, use_threads: bool = False, ncores: int = 2, **kwargs):
    """

    Downloads soil data from SSURGO (Soil Survey Geographic Database) based on lonlat coordinates.

    Args:
        ``iterable`` (iterable): An iterable containing lonlat coordinates as tuples or lists. Preferred is generator.

        ``use_threads`` (bool, optional): If True, use thread pool execution. If False, use process pool execution. Default
          is ``False``. - Ncores (int, optional): The number of CPU cores or threads to use for parallel processing. If not
          provided, it defaults to ``40%`` of available CPU cores.

    Returns:
    - a ``generator``: with dictionaries containing calculated soil profiles with the corresponding index positions based on lonlat coordinates.

    Example:

    # Example usage of download_soil_tables function
    >>> from your_module import download_soil_tables

    >>>Lonlat_coords = [(x1, y1), (x2, y2), ...]  # Replace with actual lonlat coordinates

    # Using threads for parallel processing

    >>> soil_profiles = download_soil_tables(lonlat_coords, use_threads=True, ncores=4)

        ``Kwargs``: ``func`` custom method for downloading soils
    ```

    Notes:
    - This function efficiently downloads soil data and returns calculated profiles.
    - The choice of thread or process execution can be specified with the ``use_threads`` parameter.
    - By default, the function utilizes available CPU cores or threads (40% of total) if `ncores` is not provided.
    - Progress information is displayed during execution.
    - Handle any exceptions that may occur during execution to avoid aborting the whole download

    """
    func = kwargs.get('func', None)
    progress_msg = 'Downloading soil profile: '
    Ncores = ncores
    if Ncores:
        ncores_2use = Ncores
    else:
        ncores_2use = int(cpu_count() * 0.50)
    worker = func or download_soil_table
    return custom_parallel(worker, iterable, ncores=ncores_2use, progress_message=progress_msg,
                           use_threads=use_threads)


if __name__ == '__main__':
    from examples import fnn  # the function should not be on the main for some reason

    lp = [(-92.70166631, 42.26139442), (-92.69581474, 42.26436962), (-92.64634469, 42.33703225)]
    gen_d = (i for i in range(100000))
    lm = custom_parallel(fnn, range(10000), use_thread=True, ncores=4)
    # lm2 = custom_parallel(fnn, gen_d, use_thread=True, ncores=10)
    # with a custom message
    lm = custom_parallel(fnn, range(100000), use_thread=True, ncores=4, void=True,progress_message="running function: ")
    # simple example

   # ap = [i for i in lm]



# apsim.py
import argparse
import requests

SERVER = "http://127.0.0.1:8000"

def load_model(name):
    r = requests.post(f"{SERVER}/load", json={"name": name})
    print(r.json()["message"])

def run_model():
    r = requests.post(f"{SERVER}/run")
    print(r.json()["message"])

def inspect_model():
    r = requests.get(f"{SERVER}/inspect")
    print("Model State:")
    for k, v in r.json().items():
        print(f"  {k}: {v}")
def inspect():
    r = requests.post(f"{SERVER}/inspect")
    print(r.json()["message"])


def main():
    parser = argparse.ArgumentParser(prog="apsim", description="APSIM CLI")
    subparsers = parser.add_subparsers(dest="command")

    load = subparsers.add_parser("load")

    load.add_argument("name")

    subparsers.add_parser("run")
    subparsers.add_parser("inspect")

    args = parser.parse_args()

    if args.command == "load":
        load_model(args.name)
    elif args.command == "run":
        run_model()
    elif args.command == "inspect":
        inspect_model()
    elif args.command == "ins":
        inspect()
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



import os
import sys as system
import pythonnet
from apsimNGpy.core import config
aPSim_PATH = config.get_apsim_bin_path()

def start_pythonnet():
    try:
        if pythonnet.get_runtime_info() is None:
            return pythonnet.load("coreclr")
    except:
        print("dotnet not found, trying alternate runtime")
        return pythonnet.load()


def load_pythonnet():
    """
    A method for loading Python for .NET (pythonnet) and APSIM models.

    This class provides a callable method for initializing the Python for .NET (pythonnet) runtime and loading APSIM models.
    Initialize the Python for .NET (pythonnet) runtime and load APSIM models.

        This method attempts to load the 'coreclr' runtime, and if not found, falls back to an alternate runtime.
        It also sets the APSIM binary path, adds necessary references, and returns a reference to the loaded APSIM models.

        Returns:
        -------
        lm: Reference to the loaded APSIM models

        Raises:
        ------
        KeyError: If APSIM path is not found in the system environmental variable.
        ValueError: If the provided APSIM path is invalid.

        Notes:
        It raises a KeyError if APSIM path is not found. Please edit the system environmental variable on your computer.
    Attributes:
    ----------
    None
    """

    _aPSim_Path = aPSim_PATH
    start_pythonnet()
    # use get because it does not raise key error. it returns none if not found
    aPSim_path = aPSim_PATH
    if not aPSim_path:
        raise KeyError("APSIM is not loaded in the system environmental variable")
    if 'bin' not in aPSim_path:
        aPSim_path = os.path.join(aPSim_path, 'bin')
    system.path.append(aPSim_path)
    import clr
    start_pythonnet()
    SYSTEM = clr.AddReference("System")

    MMODELSS = clr.AddReference("Models")
    apsimNG = clr.AddReference('ApsimNG')

    # return lm, sys, pythonnet.get_runtime_info()


load_pythonnet()

# Example usage:
if __name__ == '__main__':
    loader = load_pythonnet()
    loaded_models = loader
    # try importing the C# models and see if the process is successful



import argparse
import sys
from os import path

def print_msg(msg, normal=True):
    """Print messages with color."""
    color = "\033[96m" if normal else "\033[91m"
    print(f'  {color}{msg}\033[0m')

class ColoredHelpFormatter(argparse.HelpFormatter):
    """Custom argparse formatter with colored section headings."""
    def start_section(self, heading):
        heading = f"\033[91m{heading}\033[0m"
        super().start_section(heading)

def apsim_bin_path():
    parser = argparse.ArgumentParser(
        description='Manage APSIM NG bin path using apsimNGpy config.',
        formatter_class=ColoredHelpFormatter
    )

    parser.add_argument(
        '-u', '--update', type=str,
        help='Update APSIM bin path using apsimNGpy.config.set_apsim_bin_path.'
    )
    parser.add_argument(
        "-s", "--show_bin_path",
        action="store_true",
        help='Show current APSIM bin path.'
    )
    parser.add_argument(
        "-v", "--version",
        action="store_true",
        help="Show current APSIM NG version."
    )
    parser.add_argument(
        "-a", "--auto_search",
        action="store_true",
        help="Search and display available APSIM NG bin path(s)."
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(0)

    args = parser.parse_args()

    if args.update:
        if not path.exists(args.update):
            print_msg(f"{args.update} does not exist", normal=False)
            sys.exit(1)
        from apsimNGpy.core.config import set_apsim_bin_path
        set_apsim_bin_path(args.update)
        print_msg(f"Bin path updated to: {args.update}")
        sys.exit(0)

    if args.show_bin_path:
        from apsimNGpy.core.config import get_apsim_bin_path
        current_path = get_apsim_bin_path()
        print_msg(current_path)
        sys.exit(0)

    if args.auto_search:
        from apsimNGpy.core.config import auto_detect_apsim_bin_path
        auto = auto_detect_apsim_bin_path()
        if not auto:
            print_msg("No bin path detected", normal=False)
        else:
            print_msg(f"Detected: {auto}")
        sys.exit(0)

    if args.version:
        from apsimNGpy.core.runner import get_apsim_version
        v = get_apsim_version()
        print_msg(f"APSIM version: {v}")
        sys.exit(0)

if __name__ == '__main__':
    apsim_bin_path()



import os
import unittest
from unittest.mock import patch

import pandas as pd

from apsimNGpy.core.core import CoreModel, find_model, _eval_model, Models
# Import the module where CoreModel class is defined
from apsimNGpy.core.model_loader import save_model_to_file
from apsimNGpy.tests.base_test import BaseTester, set_wd

set_wd()


class TestCoreModel(BaseTester):

    def test_run(self):

        """ Test the run method's behavior under normal conditions.
         ==========================================================="""
        with patch.object(self.test_ap_sim, '_DataStore', create=True) as mock_datastore:
            # if we just run without
            """There are several ways to check if the run was successful, here I have tried all of them"""
            self.test_ap_sim.run(report_name=['Report'])
            self.assertIsInstance(self.test_ap_sim.results, pd.DataFrame,
                                  msg="an error occurred while passing  alist "
                                      "of report names")
            # one more test
            # check if the use pass  report name as str a strict a pandas.core.frame.DataFrame' is returned
            self.test_ap_sim.run(report_name='Report', verbose=False)
            df = self.test_ap_sim.results
            self.assertIsInstance(df, pd.DataFrame)
            # check if it is not empty
            self.assertEqual(df.empty, False, msg='Model was run, but results are empty')
            self.assertTrue(self.test_ap_sim.ran_ok, f'something happened in {repr(self.test_ap_sim)}')

    #            self.assertTrue(mock_datastore.Close.called)

    def test_update_mgt(self):
        """ Test the update method's behavior under
        ==========================================="""
        Amount = 23.557777
        script_name = 'Fertilise at sowing'
        mgt_script = {'Name': script_name, 'Amount': Amount},
        self.test_ap_sim.update_mgt(management=mgt_script)
        # extract user infor
        value = self.test_ap_sim.inspect_model_parameters('Manager', simulations='Simulation',
                                                          model_name='Fertilise at sowing', parameters='Amount')
        amountIn = float(value['Amount'])
        self.assertEqual(amountIn, Amount)

    def test_find_simulations(self):
        """ Test find_simulations based on three input None, lists and string
        =====================================================================
        """
        # test None
        sim = 'Simulation'
        MSG = f'test_find_simulations failed to return requested simulation object: {sim}'
        self.assertTrue(self.test_ap_sim.find_simulations(), msg=MSG)
        # test str
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=sim), msg=MSG)
        # test tuple

        self.assertTrue(self.test_ap_sim.find_simulations(simulations=(sim,)), msg=MSG)
        # test list input
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=[sim]), msg=MSG)

    def test_get_simulated_output(self):
        """ Test load_simulated_results
        requires that the models are run first
        ================================================
        """
        self.test_ap_sim.run()
        if not self.test_ap_sim.ran_ok:
            raise unittest.SkipTest(f'skipping test_simulated_results because model did '
                                    f'not run successfully or db was deleted')
        repos = self.test_ap_sim.get_simulated_output(report_names='Report')
        msg = f"expected pd.dataframe but received {type(repos)}"
        self.assertIsInstance(repos, pd.DataFrame, msg=msg)
        # test if dict not empty
        msg = 'expected dict is empty'

    def test_get_reports(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model('Report'), list)

    def test_inspect_clock(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model('Clock'), list)

    def test_inspect_simulation(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model('Models.Core.Simulation', fullpath=False), list)

    def test_replace_soil_property_values(self):
        parameter = 'Carbon'
        param_values = [2.4, 1.4]
        self.test_ap_sim.replace_soil_property_values(parameter=parameter, param_values=param_values,
                                                      soil_child='Organic', )
        lisT = self.test_ap_sim.inspect_model_parameters(model_type='Organic', simulations='all', model_name='Organic',
                                                         parameters='Carbon')
        lisT = lisT['Carbon'].tolist()
        self.assertIsInstance(lisT, list, msg='expected a list got {}'.format(type(lisT)))
        self.assertTrue(lisT)
        self.assertIsInstance(lisT, list, msg='expected a list got {}'.format(type(lisT)))
        self.assertTrue(lisT)
        # if it was successful
        testP = self.test_ap_sim.inspect_model_parameters(model_type='Organic', model_name='Organic',
                                                          parameters='Carbon')
        testP = testP['Carbon'].tolist()
        self.assertEqual(lisT[:2], param_values,
                         msg=f'replace_soil_property_values was not successful returned {testP}\n got {param_values}')

    def test_replace_soil_properties_by_path(self):
        param_values = [1.45, 1.95]
        self.test_ap_sim.edit_model(model_type='Organic', model_name='Organic', Carbon=param_values)

    def test_update_mgt_by_path(self):
        # test when fmt  ='.'
        pathto = '.Simulations.Simulation.Field.Sow using a variable rule'
        self.test_ap_sim.update_mgt_by_path(path=pathto, Population=7.5, fmt='.')
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, msg='simulation was not ran when fmt was .')
        # test when fmt = '/'
        pathto = '/Simulations/Simulation/Field.Sow using a variable rule'
        self.test_ap_sim.update_mgt_by_path(path=pathto, Population=7.5, fmt='/')
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, msg='simulation was not ran when fmt was /')

    def test_create_experiment(self):
        """creates a factorial experiment adds a factor, then test if it runs successfully
        =====================================================================================
        """
        self.test_ap_sim.create_experiment()
        # add factor
        self.test_ap_sim.add_factor(specification="[Fertilise at sowing].Script.Amount = 0 to 200 step 20",
                                    factor_name='Nitrogen')
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, msg='after adding the experiment and factor running apsim failed')

    def test_add_crop_replacements(self):
        self.test_ap_sim.add_crop_replacements(_crop='Maize')
        xp = self.test_ap_sim.Simulations.FindInScope('Replacements')
        if xp:
            rep = True
        else:
            rep = False
        self.assertTrue(rep, msg='replacement was not successful')

    def test_replace_soils_values_by_path(self):
        node = '.Simulations.Simulation.Field.Soil.Organic'
        self.test_ap_sim.replace_soils_values_by_path(node_path=node,
                                                      indices=[0], Carbon=1.3)
        v = self.test_ap_sim.get_soil_values_by_path(node, 'Carbon')['Carbon'][0]
        self.assertEqual(v, 1.3)

    def test_add_db_table(self):
        # self.test_ap_sim.add_db_table(variable_spec=['[Clock].Today', '[Soil].Nutrient.TotalC[1]/1000 as SOC1'], rename='report2')
        self.test_ap_sim.add_db_table(
            variable_spec=['[Clock].Today', '[Soil].Nutrient.TotalC[1]/1000 as SOC1',
                           '[Maize].Grain.Total.Wt*10 as Yield'],
            rename='report4', set_event_names=['[Maize].Harvesting', '[Clock].EndOfYear'])
        # check if report is created
        _reports = self.test_ap_sim.inspect_model('Models.Report()', fullpath=False)

        assert 'report4' in _reports, 'report4 was not found in reports'
        # try running
        self.test_ap_sim.run('report4')

        self.assertTrue(self.test_ap_sim.ran_ok, msg='simulation was not ran after adding the report table report2.')

    def test_loading_defaults_with(self):
        model = CoreModel(model='Maize')
        model.run()
        self.assertTrue(model.ran_ok, 'simulation was not ran after using defualt within CoreModel class')

    def test_find_model_andeval_models(self):
        """
        Unit test for the `find_model` and `_eval_model` functions.
        Verifies that:
        - Models can be found correctly by name (case-insensitive).
        - Strings representing full module paths are evaluated correctly.
        """

        # Test finding a model by simple string name
        clok_model = find_model('clock')
        _eval_model(clok_model)  # Validate model lookup
        self.assertEqual(clok_model, Models.Clock)  # Ensure it matches the Clock model

        simulation_model = find_model('Simulation')
        _eval_model(simulation_model)  # Validate model lookup
        self.assertEqual(simulation_model, Models.Core.Simulation)  # Ensure correct match

        # Test direct evaluation of model paths as strings
        clock = _eval_model('Models.Clock')
        self.assertEqual(clock, Models.Clock)  # Confirm evaluation matches Clock model

        simulation = _eval_model('Models.Core.Simulation')
        self.assertEqual(simulation, Models.Core.Simulation)  # Confirm evaluation matches Core.Simulation model

        # Test direct evaluation of just model name
        experiment = _eval_model('Experiment')
        self.assertEqual(experiment,
                         Models.Factorial.Experiment)  # Confirm evaluation matches Models.Factorial.Experiment model

        factors = _eval_model('Factors')
        self.assertEqual(factors, Models.Factorial.Factors)  # Confirm evaluation matches Models.Factorial.Factors model

    def test_add_model_simulation(self):
        try:
            self.test_ap_sim.add_model('Simulation', adoptive_parent='Simulations', rename='soybean_replaced',
                                       source='Soybean', override=True)
            self.test_ap_sim.inspect_file()
            assert 'soybean_replaced' in self.test_ap_sim.inspect_model('Simulation',
                                                                        fullpath=False), 'testing adding simulations was not successful'
        # clean up
        finally:
            from apsimNGpy.core.core import get_or_check_model
            try:
                self.test_ap_sim.remove_model(model_type='Simulation', model_name='soybean_replaced')
            except Exception as e:
                pass

    def test_edit_model(self):
        self.test_ap_sim.edit_model(model_type='Clock', simulations="Simulation", model_name='Clock',
                                    Start='1900-01-01', End='1990-01-12')
        start = self.test_ap_sim.Start != 'unknown'
        end = self.test_ap_sim.End != 'unknown'
        self.assertTrue(start, 'Simulation start date was not successfully changed')
        self.assertTrue(end, 'Simulation end date was not successfully changed')

    def __test_edit_cultivar_edit_model_method(self):
        """
        Test the edit_cultivar requires that we have replacements in place
        @return:
        """
        from apsimNGpy.core.structure import add_crop_replacements
        add_crop_replacements(self.test_ap_sim, _crop='Maize')
        in_cul = self.test_ap_sim.extract_user_input('Sow using a variable rule')
        in_cultivar = in_cul['Simulation'].get('CultivarName')
        if not in_cultivar:
            raise unittest.SkipTest('skipping test edit cultivar because cultivar not found')
        new_juvenile = 289.777729777
        self.test_ap_sim.edit_model(model_type='Cultivar', model_name=in_cultivar,
                                    commands='[Phenology].Juvenile.Target.FixedValue', values=new_juvenile,
                                    cultivar_manager='Sow using a variable rule')

        # first we check the current '[Phenology].Juvenile.Target.FixedValue': '211'
        cp = self.test_ap_sim.read_cultivar_params(name=in_cultivar)
        com_path = '[Phenology].Juvenile.Target.FixedValue'
        juvenile_original = float(cp.get(com_path))

        self.test_ap_sim.edit_cultivar(commands=com_path, values=new_juvenile, CultivarName=in_cultivar)
        # read again
        juvenile_replacements = self.test_ap_sim.read_cultivar_params(name=in_cultivar)[com_path]
        if juvenile_replacements:
            juvenile_replacements = float(juvenile_replacements)
        self.assertEqual(new_juvenile, juvenile_replacements)
        self.test_ap_sim.run()
        # self.assertGreater(juvenile_replacements, juvenile_original, msg= 'Juvenile target value was not replaced')

    def test_inspect_model_parameters(self):

        w_file = self.test_ap_sim.inspect_model_parameters(model_type='Weather', simulations='Simulation',
                                                           model_name='Weather')
        self.assertIsInstance(w_file, str, msg='Inspect weather model parameters failed, when simulation is a str')
        # test a list of simulations
        w_file = self.test_ap_sim.inspect_model_parameters(model_type='Weather', simulations=['Simulation'],
                                                           model_name='Weather')
        self.assertIsInstance(w_file, dict, msg='Inspect weather model parameters failed, when simualtion is a list')


if __name__ == '__main__':
    ...
    unittest.main()



from apsimNGpy.manager.soilmanager import DownloadsurgoSoiltables, OrganiseSoilProfile
from apsimNGpy.manager.weathermanager import get_met_from_day_met
from apsimNGpy.core.apsim import ApsimModel


def initialise(model, reports):
    model = ApsimModel(model)
    model.run(report_name=reports)
    return model.results


def download_soil_table(x):
    try:
        cod = x
        table = DownloadsurgoSoiltables(cod)
        th = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]
        sp = OrganiseSoilProfile(table, thickness=20, thickness_values=th).cal_missingFromSurgo()
        return {x: sp}
    except Exception as e:
        print("Exception Type:", type(e), "has occured")
        print(repr(e))


def simulator_worker(row, dictio):
    kwargs = dictio
    report = kwargs.get('report_name')
    ID = row['ID']
    model = row['file_name']
    thi = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]
    th = kwargs.get("thickness_values", thi)
    simulator_model = ApsimModel(
        model, copy=kwargs.get('copy'), read_from_string=True, lonlat=None, thickness_values=th)
    sim_names = simulator_model.extract_simulation_name
    location = row['location']
    stat, end = kwargs.get('start'), kwargs.get('end')
    if kwargs.get('replace_weather', False):
        wname = model.strip('.apsimx') + '_w.met'
        wf = get_met_from_day_met(location, start=stat, end=end, filename=wname)
        simulator_model.replace_met_file(wf, sim_names)

    if kwargs.get("replace_soil", False):
        table = DownloadsurgoSoiltables(location)
        sp = OrganiseSoilProfile(table, thickness=20, thickness_values=th)
        sp = sp.cal_missingFromSurgo()
        simulator_model.replace_downloaded_soils(sp, sim_names)

    if kwargs.get("mgt_practices"):
        simulator_model.update_mgt(kwargs.get('mgt_practices'), sim_names)
    try:
        simulator_model.run(report_name=report)
        return simulator_model.results
    except Exception as e:
        print(type(e))
        print('+_____________________________________\n')
        print(e)



import os

from setuptools import setup, find_packages

VERSION = '0.39.1'

DESCRIPTION = 'APSIM next generation package interface'
LONG_DESCRIPTION = 'run, edit, download soils and weather and interact with the apsimx file'

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name='apsimNGpy',
    version=VERSION,
    url='https://github.com/MAGALA-RICHARD/apsimNGpy.git',
    license='MIT',
    author='Richard Magala',
    author_email='magalarich20@gmail.com',
    description=DESCRIPTION,
    long_description=readme,
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'apsim=apsimNGpy.cli.cli:main_entry_point',
            'bp=apsimNGpy.cli.set_bin:fast',
            'client=apsimNGpy.cli.client:app',
            'apsim_bin_path=apsimNGpy.cli.set_ups:apsim_bin_path',
        ],
    },

    package_data={'': ['./apsimNGpy/data/*.apsimx',
                       './apsimNGpy/*.met',
                       "./apsimNGpy/experiment/*.py",
                       './apsimNGpy/examples/*.png',
                       './apsimNGpy/images/*.png'
                       './apsimNGpy/*.ini', "./*.ini"]},
    keywords=['python', 'APSIM Next Generation', 'pythonnet', 'crop modeling'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=[
        'numpy>=1.20.0',
        'scipy>=1.7',
        'xmltodict >=0.13.0',
        'requests >=2.31.0',
        'pythonnet >= 3.0.1',
        'rasterio >=1.3.8',
        'shapely >=2.0.1',
        'clr >= 1.0.3',
        'geopandas >=0.13.2',
        'pandas >=2.0.3',
        'geopy >= 2.4.1',
        'tqdm >= 4.66.2',
        'progressbar >= 2.5',
        'joblib >= 1.3.2',
        'sqlalchemy >=2.0',
        'matplotlib',
        'psutil >=6.0.0',
        'tenacity',
        'typer',
        'summarytools>=0.3.0'

    ]
)



from .utils import collect_runfiles, load_from_numpy, upload_apsimx_file_by_pattern, assign_management_practices, make_apsimx_clones,\
    organize_crop_rotations, evaluate_crop
__all__ = ['load_from_numpy', 'collect_runfiles']


import stat
import unittest
from pathlib import Path

from apsimNGpy.tests import (unit_tests_core,
                             unit_test_apsim,
                             test_config,
                             test_edit_model,
                             test_weathermanager)


def clean_up():
    sc = Path('../scratch')
    for path in sc.rglob("temp_*"):
        try:
            # Ensure file is writable (especially on Windows)
            path.chmod(stat.S_IWRITE)
            path.unlink(missing_ok=True)
        except PermissionError:
            ...
        except Exception as e:
            ...


# Create a test suite combining all the test cases
loader = unittest.TestLoader()
suite = unittest.TestSuite()

suite.addTests(loader.loadTestsFromModule(test_config))
suite.addTests(loader.loadTestsFromModule(test_weathermanager))
suite.addTests(loader.loadTestsFromModule(unit_tests_core))
suite.addTests(loader.loadTestsFromModule(unit_test_apsim))
suite.addTests(loader.loadTestsFromModule(test_edit_model))


def run_suite(verbosity_level=2):
    try:
        runner = unittest.TextTestRunner(verbosity=verbosity_level)
        runner.run(suite)
    finally:
        clean_up()


if __name__ == '__main__':

        # Run the test suite
        clean_up()
        run_suite(3)





# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'apsimNGpy'
copyright = '2025, richard magala'
author = 'richard magala'
release = '0.32'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [

]

templates_path = ['_templates']
exclude_patterns = ['sphinx.ext.duration',
'sphinx.ext.napoleon',
                    'sphinx.ext.doctest',
                    'sphinx.ext.autodoc',
                    'sphinx.ext.autosummary',
                    'sphinx.ext.githubpages',
                    "sphinx.ext.intersphinx",
                    ]

intersphinx_mapping = {
    "rtd": ("https://docs.readthedocs.io/en/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]
autodoc_mock_imports = []  # Add external modules to mock if they are not installed

autosummary_generate = True  # Automatically create stub pages for modules
# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# -- Options for EPUB output
epub_show_urls = "footnote"

html_theme = "sphinx_rtd_theme"

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", '.apsimx', '.db']

html_static_path = ['_static']

import os, sys
sys.path.insert(0, os.path.abspath('../'))



from shapely import wkt
import pandas as pd
import os


def convert_df_to_gdf(df, CRS):
    """
    Converts a pandas DataFrame to a GeoPandas GeoDataFrame with specified CRS.

    This function assumes that the input DataFrame contains a column named 'geometry'
    with WKT (Well-Known Text) representations of geometric objects. It converts these
    textual geometry representations into GeoPandas geometry objects and sets the
    coordinate reference system (CRS) for the resulting GeoDataFrame.

    Parameters:
    - df (pandas.DataFrame): The input DataFrame to convert. Must include a 'geometry'
      column with WKT representations of the geometries.
    - CRS (str or dict): The coordinate reference system to set for the GeoDataFrame.
      Can be specified as a Proj4 string, a dictionary of Proj parameters, an EPSG
      code, or a WKT string.

    Returns:
    - gpd.GeoDataFrame: A GeoDataFrame created from the input DataFrame with geometries
      converted from WKT format and the specified CRS.

    Example:
    ```
    import pandas as pd
    import geopandas as gpd
    from shapely import wkt

    # Sample DataFrame with WKT geometries
    data = {'geometry': ['POINT (12 34)', 'LINESTRING (0 0, 1 1, 2 1, 2 2)'],
            'value': [1, 2]}
    df = pd.DataFrame(data)

    # Convert DataFrame to GeoDataFrame with an EPSG code for CRS
    gdf = convert_df_to_gdf(df, CRS='EPSG:4326')
    ```

    Note:
    The 'wkt' module from 'shapely' needs to be imported to convert geometries from WKT.
    """
    df['geometry'] = df['geometry'].apply(wkt.loads)
    gdf = gpd.GeoDataFrame(df, crs=CRS)
    return gdf


def match_crop(abb, add_wheat=None):
    """
    Converts agricultural abbreviations to a string of crop names, with an option to insert 'Wheat' at specified 
    intervals. The function interprets abbreviations where 'B' stands for Soybean, 'C' for Maize, 'P' for Pasture, 
    'U' for Developed areas, 'T' for Wetlands, 'F' for Forest, and 'R' for Other crops.

    This function takes an abbreviation string representing a sequence of crops and optionally inserts 'Wheat' into 
    the sequence at predetermined positions. It handles special cases where the abbreviation matches certain 
    patterns, returning None for these cases.

    Parameters:
    - abb (str): The abbreviation string representing a sequence of crops. Each character stands for a specific crop:
        'C' for Maize,
        'B' for Soybean,
        'W' for Wheat,
        'P' for Tropical Pasture.
      Characters 'U', 'R', 'T', and 'F' are ignored and removed from the abbreviation.
    - add_wheat (bool, optional): If True, 'Wheat' is inserted at positions 1, 3, 5, 7, 9, and 11 in the sequence. Defaults to None, which means 'Wheat' is not added.

    Returns:
    - str or None: A comma-separated string of crop names based on the abbreviation. Returns None if the abbreviation matches special cases 'TTTTTT' or 'UUUUUUU'.

    Example:
    ```
    # Convert abbreviation to crop names with additional 'Wheat'
    crop_sequence = match_crop("CBWP", add_wheat=True)
    print(crop_sequence)
    # Output: "Maize,Wheat,Soybean,Wheat,Wheat,Tropical Pasture"

    # Convert abbreviation without adding 'Wheat'
    crop_sequence = match_crop("CBWP")
    print(crop_sequence)
    # Output: "Maize,Soybean,Wheat,Tropical Pasture"
    ```

    Note:
    - The function returns None for special case abbreviations 'TTTTTT' and 'UUUUUUU' as an indication of non-standard sequences that should not be ran_ok.
    - The insertion of 'Wheat' at specified positions assumes the sequence is long enough to accommodate these insertions. If the sequence is shorter, 'Wheat' will only be inserted up to the length of the sequence.
    """
    # Reserve the original abbreviation
    reserve = abb

    # Check if abbreviation does not match special cases and contains certain characters
    if abb not in ['TTTTTT', 'UUUUUUU'] and ('C' in abb or 'B' in abb or 'W' in abb or 'P' in abb):
        # Remove specified characters from the abbreviation
        work = abb.replace("U", "").replace("R", "").replace("T", "").replace("F", "")

        # Replace abbreviations with crop names, adding commas between them
        rotation = work.replace("C", "Maize,").replace("B", "Soybean,").replace("W", "Wheat").replace('P',
                                                                                                      'TropicalPasture').rstrip(
            ",")
        rot2 = rotation.split(",")

        if add_wheat:
            # Insert "Wheat" at specified positions if add_wheat is True
            for i in [1, 3, 5, 7, 9, 11]:
                if i < len(rot2):
                    rot2.insert(i, "Wheat")
            crop = ",".join(rot2)
        else:
            # Join the rotation without additional "Wheat"
            crop = ",".join(rot2)
    else:
        # If abbreviation matches special cases, return it as is
        crop = None

    return crop


from functools import singledispatch


@singledispatch
def append(obj, x):
    print("Unsupported child")


@append.register
def _(obj: list, x: list):
    return obj + x


@append.register
def _(obj: set, x: set):
    return obj.union(x)


@append.register
def _(obj: str, x: str):
    return obj + x


print(append([1, 2, 3], [4, 5]))
print(append({1, 2, 3}, {4, 5}))
print(append("1 2 3", " 4 5"), "\n")

append(2, 3)

if __name__ == "__main__":
    al = match_crop('CBBBBC')

    print(al)
    csv = r'D:\ACPd\Long deek401 simulations\lu_table.csv'
    df = pd.read_csv(csv)
    col = df.CropRotatn

    df['rotation_cover_crop'] = df['CropRotatn'].apply(lambda x: match_crop(x, add_wheat=True))
    df['rotations'] = df['CropRotatn'].apply(lambda x: match_crop(x, add_wheat=False))

    print(df)
    df.to_csv("fin.csv")
    # os.startfile('fin.csv')
    from functools import total_ordering


   


import os
import shutil
from pathlib import Path
import copy
from dataclasses import dataclass
import inspect
from typing import Any
from abc import ABC, abstractmethod
import numpy as np
from scipy.optimize import minimize
from apsimNGpy.optimizer.variables import auto_guess
from apsimNGpy.core.apsim import ApsimModel
from typing import Iterable


@dataclass(slots=True)
class Solvers:
    L_BFGS_B = 'L-BFGS-B'
    BFGS = 'BFGS'
    Powell = 'Powell'
    Newton_CG = 'Newton-CG'
    lm = 'lm'
    Nelder_Mead = 'Nelder-Mead'
    SLSQP = 'SLSQP'


def _fun_inspector(fun):
    assert callable(fun), 'method supplied is not a callable object'
    sign = inspect.signature(fun)
    params = [i.name for i in sign.parameters.values()]
    return params


class Problem(ABC):

    def __init__(self):

        self.WS = None
        self.user_ws = None
        self.controls = None
        self.n_workers = 1
        self.evaluation_sign = None
        self.out_path = None
        self.model = None
        self.options = None
        self.func = None
        self.observed = None

    def set_up_data(self, model: str, func: callable, ws: os.PathLike, observed_values: np.ndarray = None,
                    options: dict = None, n_workers=1):
        """
        model: APSIM model file, apsimNGpy object, apsim file str or dict we want to use in the minimization func: an
        evaluation function, which is callable.
        This is something you should write for your self.
        A typical evaluation function takes on the functional signature func(apsimNGpy.CoreModel.model, *args) Additional
        arguments can be passed as a tuple.
        Example of this could be observed _variables to be compared with the
        predicted, where a loss function like rmse errors can be computed.
        In case of extra argument, these should be
        passed via options e.g., options ={args: None}
         @return: an instance of problem class object.
        """
        # re-initialize all lists carrying the data
        self.reset_data()
        self.set_ws(ws)
        self.options = options
        self.model = model
        self.n_workers = n_workers
        self.observed = observed_values
        self.user_ws = ws

        self.evaluation_sign = _fun_inspector(func)

        # make sure the user is consistent with extra argument
        if 'args' in self.evaluation_sign and not self.options.get('args'):
            raise ValueError("function evaluator has extra arguments not specified")
        self.func = func

        return self

    def add_control_variable(self, control: Any):
        self.controls.append(control)
    def add_control(self, model_type, model_name, simulation_name, variable_name):
        self.controls.append((model_type, model_name, simulation_name, variable_name))
    def update_predictors(self, x: tuple):
        if len(self.controls) < 1:
            raise ValueError("No control _variables added yet")
        vals = '_'.join(str(v) for v in x)
        # we want a unique out_path name for parallel processing
        if self.n_workers > 1:
            new_path = self.WS.joinpath(f"{vals}.apsimx")

        if not isinstance(self.model, Iterable):
            model_s = [self.model]
        else:
            model_s = self.model
        MODELs = []
        for model_x in model_s:
            model = ApsimModel(model_x)
            try:

                for predictor, x_to_fill in zip(self.controls, x):
                    # update main param for the function updator
                    predictor.params[predictor.main_param] = x_to_fill

                    getattr(model, predictor.updater)(**predictor.params)
                    ui = model.extract_user_input('SowMaize')

                    # xit loop and run
                    # model.run(report_name='MaizeR')
                    MODELs.append(model)
            finally:
                pass

        try:
            if 'args' in self.evaluation_sign:
                ans = self.func(MODELs, *self.options['args'])
            else:
                ans = self.func(MODELs)

            return ans
        finally:
            for model in MODELs:
                try:
                    datastore = model.datastore

                    os.remove(model.path)
                    print(f"Removed: {datastore}")
                    os.remove(datastore)
                    del model
                except (FileNotFoundError, PermissionError) as e:

                    pass

    def minimize_problem(self, **kwargs):

        """

        Minimizes the defined problem using scipy.optimize.minimize.
        kwargs: key word arguments as defined by the scipy minimize method
        see scipy manual for each method https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html

        Returns:
        Optimization result.

        """
        # freeze the data before optimization to make it immutable
        self._freeze_data()

        minim = minimize(self.update_predictors, **kwargs)
        labels = [i.label for i in self.controls]
        ap = dict(zip(labels, minim.x))
        setattr(minim, 'x_vars', ap)
        return minim

    def reset_data(self):
        """
        Resets the data for the optimizer. You don't need to call set up in the data again by calling the
        set_up_data_method in case your setup does need to change. After resting, you have to add the control vars
        @return:
        """
        self.controls = []

        return self

    def set_ws(self, ws):
        self.WS = Path(ws).joinpath('out_files')
        if self.WS.exists():
            try:
                shutil.rmtree(self.WS)
            except (FileNotFoundError, PermissionError):
                ...
        self.WS.mkdir(exist_ok=True)

    def _freeze_data(self):
        """
        # make the data unchangeable after editing.
        # Please note that this only works during execution.
        # I expect some leakages before
         """
        self.controls = tuple([i for i in self.controls])


# initialized before it is needed
SingleProblem = Problem()



import unittest
from pandas import DataFrame
from apsimNGpy.tests.base_test import BaseTester, set_wd

set_wd()


class TestExperiment(BaseTester):

    def test_add_factor_with_permutes(self):
        self.test_ap_sim.create_experiment(permutation=True)
        # experiment is initialized in base_tester class SetUp
        self.test_ap_sim.add_factor(specification="[Fertilise at sowing].Script.Amount = 0 to 200 step 50",
                                   factor_name='Nitrogen')
        self.test_ap_sim.run()
        self.assertIsInstance(self.test_ap_sim.results, DataFrame)
        if isinstance(self.test_ap_sim.results, DataFrame):
            self.assertFalse(self.test_ap_sim.results.empty, msg='results are empty')
    def test_add_factor_without_permutes(self):
        self.test_ap_sim.create_experiment(permutation=False)
        # experiment is initialized in base_tester class SetUp
        self.test_ap_sim.add_factor(specification="[Fertilise at sowing].Script.Amount = 0 to 200 step 50",
                                   factor_name='Nitrogen')
        self.test_ap_sim.run()
        self.assertIsInstance(self.test_ap_sim.results, DataFrame)
        if isinstance(self.test_ap_sim.results, DataFrame):
            self.assertFalse(self.test_ap_sim.results.empty, msg='results are empty')


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase


class TestCoreModel(TestCase):
    def test_check_model(self):
        self.fail()

    def test__remove_related_files(self):
        self.fail()

    def test_simulations(self):
        self.fail()

    def test_simulation_names(self):
        self.fail()

    def test_run(self):
        self.fail()

    def test_simulated_results(self):
        self.fail()



import logging
import os
import traceback
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pygments.lexers import nit
import copy
from config import *

info = 'info'
error = 'error'
SoilTable = 'SERF'

from apsimNGpy.core.apsim import ApsimModel

run = True
if __name__ == '__main__' and run:
    apsi_m = wdi.path(name='apsim/baseC.apsimx')
    directory = wdi.path(name='NWREC_edited')
    wdi.mkdir('NWREC_edited')
    model = ApsimModel(apsi_m)
    plot_path = {}
    Crops = 'Maize'
    nw_csv = pd.read_csv(wdi.path(name='observed_soc.csv'))


    # NREW N fertilizer was 246
    def _mgt(crop, depth=350, n_amount=200):
        print(crop)
        if 'Soybean' in crop:
            n_amount = 170
        sowManager = {"Name": 'MaizeNitrogenManager', 'Amount': n_amount}
        return [{"Name": 'Simple Rotation', 'Crops': crop},
                {'Name': 'Tillage', 'Depth': depth, 'AppplyTillageOption': 'no', 'Fraction': 0.65}, sowManager]


    apsim_edited_files = joblib.load(os.path.join(directory, 'edited_files.joblib'))
    scratch = Path.cwd() / 'scratch'
    scratch.mkdir(exist_ok=True)
    DATA = [ApsimModel(filename).run(['MaizeR', "Carbon", "Annual"]).results for filename in
            apsim_edited_files.values()]
    for data, plot in zip(DATA, apsim_edited_files.keys()):
        data['plotid'] = plot
    predicted = pd.concat(DATA)
    mYield = predicted[~np.isnan(predicted['Maize_Yield'])].copy()
    mYield.dropna(subset='Year', inplace=True)
    mYield.rename(columns={'Year': 'year'}, inplace=True)
    socP = predicted[~np.isnan(predicted['SOC1'])].copy()
    socP['year'] = pd.DatetimeIndex(socP.Date).year.to_list()
    all_df = pd.merge([socP, nw_csv], how=inner, on=['plotid', 'year'])
    df_merged = nw_csv.drop_duplicates(subset=['plotid', 'depth', 'year'])

    data = df_merged.copy()

    data['socx'] = data['SOC1'] / (data['BD'] * data['depth'])
    data['Carbon'] = data['Carbon'] / 10
    # data.to_csv(wdi.path(name='carbon.csv'))
    dat = data.copy()
    # data = dat[dat['year']==2014.0].copy()
    val = validate(actual=data.Carbon, predicted=data.socx)
    ev_all = val.evaluate_all(verbose=True)
    X = data[['Carbon']].values
    y = data['socx'].values
    plot_reg_fit(X, y, data=data)

    C_yield = load_AGRON(Yield, XLs[SoilTable])
    mY = validate(actual=C_yield.obs / 1000, predicted=C_yield.Maize_Yield / 1000)
    mY.evaluate_all(verbose=True)
    from apsimNGpy.visual.visual import quick_plot

    print(C_yield)
    plot_po(xx=data.year, predicted=data.socx, observed=data.Carbon, anotate_with=data.plotid)
    data.to_csv(wdi.path(name='ccs.csv'))



import matplotlib.pyplot as plt


def plot_data(x, y, plot_type='line', xlabel = 'X-axis', ylabel = "Y-axis"):
    """
    Plot data points.

    Parameters:
    x (list or array-like): X-axis data.
    y (list or array-like): Y-axis data.
    plot_type (str): Type of plot ('line', 'scatter', 'bar', 'hist', 'box', 'pie').
    """
    if plot_type == 'line':
        plt.plot(x, y, label=ylabel)
    elif plot_type == 'scatter':
        plt.scatter(x, y, label=ylabel)
    elif plot_type == 'bar':
        plt.bar(x, y, label='Bar Plot')
    elif plot_type == 'hist':
        plt.hist(y, bins=10, label='Histogram')
    elif plot_type == 'box':
        plt.boxplot(y, vert=False, labels=['Box Plot'])
    elif plot_type == 'pie':
        plt.pie(y, labels=x, autopct='%1.1f%%', startangle=90)
    else:
        print("Invalid plot child. Available types: 'line', 'scatter', 'bar', 'hist', 'box', 'pie'")
        return

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(f'{plot_type.capitalize()} Plot')
    plt.legend()
    plt.show()
    plt.savefig("plotted.jpeg")


if __name__ == "__main__":
    # Example usage:
    x_data = [1, 2, 3, 4, 5]
    y_data = [10, 7, 5, 2, 8]
    plot_type = 'line'  # Change this to the desired plot child
    plot_data(x_data, y_data, plot_type)



from dataclasses import dataclass, field
from apsimNGpy.core.apsim import ApsimModel
from scipy.optimize import minimize
import numpy as np
from functools import reduce, lru_cache
from collections import OrderedDict
from functools import wraps
from collections import OrderedDict


class SimpleCache:
    def __init__(self, maxsize=None):
        self.cache = OrderedDict()
        self.maxsize = maxsize

    def cache_res(self, *args, out=None, **kwargs):
        key = (args, tuple(sorted(kwargs.items())))

        if key in self.cache:
            return self.cache[key]

        self.cache[key] = out

        if self.maxsize is not None and len(self.cache) > self.maxsize:
            self.cache.popitem(last=False)  # FIFO: remove oldest

    def get(self, *args, **kwargs):
        key = (args, tuple(sorted(kwargs.items())))
        return self.cache.get(key, None)

    def clear(self):
        self.cache.clear()


results = {}
@dataclass
class Problem:
    apsim: ApsimModel
    simulation: str
    controls: list = field(default_factory=list)
    labels: list = field(default_factory=list)
    cache: bool = field(default_factory=bool)
    cache_size: int = field(default_factory=int)

    def __post_init__(self):
        if self.cache:
            cache_res = SimpleCache(maxsize=self.cache_size)

    def add_control(self, model_type, model_name, parameter_name, label=None):
        self.controls.append({
            "model_type": model_type,
            "model_name": model_name,
            "parameter_name": parameter_name,
            "label": label or f"{model_type}_{model_name}_{parameter_name}"
        })

    def setUP(self, x):
        sim = self.apsim.inspect_model('Simulation', fullpath=False)[0]
        edit = self.apsim.edit_model
        for i, varR in enumerate(self.controls):
            print(x)
            edit(
                model_type=varR['model_type'],
                simulations='Simulation',
                model_name=varR['model_name'],
                cacheit=True,
                **{varR['parameter_name']: x[i]}
            )


    def evaluate(self, x):
        x = tuple([np.round(i) for i in x])
        if results.get(x, None):
            return results[x]
        self.setUP(x)
        result = self.apsim.run(verbose=False).results
        emissions = result.Yield  # placeholder
        result=  10000 - emissions.mean()
        results[x] = result
        return result


    def minimize_problem(self, **kwargs):


        x0 = kwargs.pop("x0", [1] * len(self.controls))  # Starting guess
        result = minimize(self.evaluate, x0=x0, **kwargs)

        labels = [c['label'] for c in self.controls]
        result.x_vars = dict(zip(labels, result.x))
        return result

if __name__ == "__main__":
    model = ApsimModel('Maize')
    prob = Problem(model, "Simulation", cache_size=200, cache=True)
    prob.add_control('Manager', "Sow using a variable rule", 'Population', 'population')
    prob.add_control('Manager', "Fertilise at sowing", 'Amount', 'Nitrogen')
    ans  =prob.minimize_problem(x0=[1, 0], method  ='Powell', bounds =[(1,12), (0, 208)])
    print(ans)


from fastapi import FastAPI, HTTPException, Query
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.core.config import (
    set_apsim_bin_path,
    get_apsim_bin_path,
    auto_detect_apsim_bin_path,
)
from apsimNGpy.core.runner import get_apsim_version
from os import path
from pydantic import BaseModel
import os
app = FastAPI(title="APSIM Binary Manager API")


@app.get("/set-path")
def set_bin_path(update: str = Query(..., description="Path to set as APSIM bin path")):
    if not path.exists(update):
        raise HTTPException(status_code=400, detail=f"Path '{update}' does not exist.")
    set_apsim_bin_path(update)
    return {"message": f"Updated APSIM bin path to: {update}"}


@app.get("/get-path")
def show_bin_path():
    cbp = get_apsim_bin_path()
    return {"current_bin_path": cbp}


@app.get("/auto-detect")
def auto_search_path():
    auto = auto_detect_apsim_bin_path()
    if not auto:
        raise HTTPException(status_code=404, detail="No APSIM bin path detected.")
    return {"detected_path": auto}


@app.get("/version")
def get_version():
    version = get_apsim_version()
    return {"apsim_version": version}

model = None

class LoadRequest(BaseModel):
    name: str

@app.post("/load")
def load_model(req: LoadRequest):
    global model
    model = ApsimModel(req.name)
    return {"message": f"Model '{req.name}' loaded."}
@app.post("/inspect")
def inspect(_model:str):
    insp = model.inspect_model(_model)
    return {"model_inspect": insp}
@app.post("/run")
def run_model():
    if not model:
        raise HTTPException(status_code=400, detail="Model not loaded.")
    model.run(verbose=True)
    model.results.to_csv('m.csv')
    os.startfile('m.csv')
    return {"message": "Model is running."}

def fast():
    import uvicorn
    uvicorn.run("apsimNGpy.cli.set_bin:app", host="127.0.0.1", port=8000, reload=True)





import argparse
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.settings import logger
from apsimNGpy.cli.model_desc import extract_models
import asyncio
from functools import lru_cache
from apsimNGpy.cli._args import parser
from apsimNGpy.cli.commands import (ApsimModel, os, sys, fetch_weather_data, replace_soil_from_web, create_experiment,replace_soil_data,pd,
                           np, print_msg, run_apsim_model, save_results, change_mgt )

async def main():
    # Create argument parser

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    # Parse arguments
    args = parser.parse_args()
    logger.info(f"Commands summary: '{args}'")
    if len(sys.argv) > 3:
        print_msg(msg='Processing in puts. Please wait!', normal=False)
    if  args.get_web_data !='no' and not args.lonlat:
        raise ValueError("attempting to fetch soil without supplying lonlat values")

    wd = args.wd or os.getcwd()
    os.makedirs(wd, exist_ok=True)

    met_form_loc = None

    if args.lonlat is not None:
        lonlat_tuple = tuple(map(float, args.lonlat.split(',')))
    else:
        lonlat_tuple = None

    if args.get_web_data == 'both' or args.get_web_data == 'w':

        met_form_loc = await fetch_weather_data(lonlat_tuple)




    file_name = args.save or f"out_{args.model.strip('.apsimx')}.csv"
    if args.lonlat and args.get_web_data == 's' or args.get_web_data == 'both':
        soilP = replace_soil_from_web(lonlat_tuple)

    if args.model.endswith('.apsimx'):
        model = ApsimModel(args.model, args.out, thickness_values=settings.SOIL_THICKNESS)
    else:
        model = load_default_simulations(crop=args.model, simulations_object=True, set_wd=wd)

    if args.inspect:
        print()
        # inspect returns after excecutions
        if args.inspect != 'file':
            model_type = args.inspect
            print_msg(model.inspect_model(model_type=model_type))

        else:
            model.inspect_file()
        print()
        return

    model = await asyncio.to_thread(change_mgt, model, args)

    await asyncio.to_thread(replace_soil_data, model, args)
    met_data = args.met_file or met_form_loc
    if met_data:

        await asyncio.to_thread(model.replace_met_file, weather_file=met_data, simulations=args.simulation)
        msg = f'Successfully updated weather file with {met_data}'
        if args.lonlat:
            msg += f' from location: {args.lonlat}'
        print_msg(msg)

    if args.lonlat and args.get_web_data == 's' or args.get_web_data == 'both':
        await asyncio.to_thread(model.replace_downloaded_soils,soilP, model.simulation_names)
    if args.experiment:
        await create_experiment(model, args.experiment)
    # Run APSIM asynchronously
    model = await run_apsim_model(model, report_name=args.table)

    df = model.results

    if isinstance(df, pd.DataFrame):
        await save_results(df, file_name)
        numeric_df = df.select_dtypes(include=np.number)
        stati = getattr(numeric_df, args.aggfunc)()

        print_msg(stati)
    if args.save_model:
        model.save(args.save_model)
    if args.preview =='yes':
        model.preview_simulation()
    sys.exit(1)


@timer
def main_entry_point() -> None:
    asyncio.run(main())


# Run asyncio event loop
if __name__ == "__main__":

    main_entry_point()
    #-m maize -sf 'm.csv' --organic "node_path=.Simulations.Simulation.Field.Soil, Carbon=[1.2]"
